/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution. */

/* iem-gui-lib written by Thomas Musil (c) IEM KUG Graz Austria 2000-2001 */


/* ---------- my_canvas  my gui-canvas for a window ---------------- */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "m_imp.h"
#include "g_canvas.h"
#include "t_tk.h"
#include "g_7_guis.h"

t_widgetbehavior my_canvas_widgetbehavior;
static t_class *my_canvas_class;

typedef struct _my_canvas
{
     t_object x_obj;
     t_atom   x_at[3];
     t_glist  *x_glist;
     int      x_size;
     int      x_w;
     int      x_h;
     char     x_font[16];
     int      x_fontsize;
     int      x_ldx;
     int      x_ldy;
     int      x_bcol;
     int      x_lcol;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_selected;
     int      x_shiftdown;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_my_canvas;

/* widget helper functions */

void my_canvas_draw(t_my_canvas *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(firsttime)
    {
        sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -outline #%6.6x -tags %xRECT\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h,
		 x->x_bcol, x->x_bcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -outline #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size,
		 x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
    }
    else
    {
        sys_vgui(".x%x.c coords %xRECT %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos, x->x_obj.te_xpos + x->x_w,
		 x->x_obj.te_ypos + x->x_h);
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx,
		     x->x_obj.te_ypos+x->x_ldy);
    }
}

void my_canvas_erase(t_my_canvas* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xRECT\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void my_canvas_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_my_canvas *x = (t_my_canvas *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void my_canvas_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_my_canvas *x = (t_my_canvas *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    my_canvas_draw(x, glist, 0);
}

static void my_canvas_select(t_gobj *z, t_glist *glist, int selected)
{
    t_my_canvas *x = (t_my_canvas *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
        sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, x->x_bcol);
    }
}

static void my_canvas_delete(t_gobj *z, t_glist *glist)
{
}

static void my_canvas_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_my_canvas *x = (t_my_canvas *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 my_canvas_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 my_canvas_erase(x, glist);
         rtext_free(y);
    }
}

static void my_canvas_save(t_gobj *z, t_binbuf *b)
{
    t_my_canvas *x = (t_my_canvas *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiisssiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("my_canvas"), x->x_size, x->x_w, x->x_h,
		x->x_snd, x->x_rcv, x->x_lab, x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8, x->x_fontsize,
		bcol, lcol);
    binbuf_addv(b, ";");
}

static void my_canvas_properties(t_gobj *z, t_glist *owner)
{
    t_my_canvas *x = (t_my_canvas *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s MY_CANVAS \
	    ------selectable_dimensions(pix):------ %g %g size: 0.0 0.0 empty \
	    ------visible_rectangle(pix)(pix):------ %g width: %g height: %g \
	    %g empty empty %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_size, 1.0,
	    (float)x->x_w, (float)x->x_h, 0.0,/*no_schedule*/
	    -1.0, -1.0, -1.0,/*no linlog, no init, no multi*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no frontcolor*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void my_canvas_get_pos(t_my_canvas *x)
{
    if(x->x_snd_able && x->x_snd->s_thing)
    {
	x->x_at[0].a_w.w_float = x->x_obj.te_xpos;
	x->x_at[1].a_w.w_float = x->x_obj.te_ypos;
	pd_list(x->x_snd->s_thing, &s_list, 2, x->x_at);
    }
}

static void my_canvas_dialog(t_my_canvas *x, t_symbol *s, int argc, t_atom *argv)
{
    char font[20];
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int a = (int)atom_getintarg(0, argc, argv);
    int w = (int)atom_getintarg(2, argc, argv);
    int h = (int)atom_getintarg(3, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    if(a < 1)
	a = 1;
    x->x_size = a;
    if(w < 1)
	w = 1;
    x->x_w = w;
    if(h < 1)
	h = 1;
    x->x_h = h;
    x->x_lcol = lcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    sys_vgui(".x%x.c itemconfigure %xRECT -fill #%6.6x -outline #%6.6x\n", canvas, x,
	     x->x_bcol, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x,
	         x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_bcol);
    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize, x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize, x->x_lcol, x->x_lab->s_name);
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_size(t_my_canvas *x, t_float size)
{
    if(size < 1.0)
        size = 1.0;

    x->x_size = (int)size;
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_delta(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_pos(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_vis_size(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    int i;

    i = (int)atom_getintarg(0, ac, av);
    if(i < 1)
	i = 1;
    x->x_w = i;
    i = (int)atom_getintarg(1, ac, av);
    if(i < 1)
	i = 1;
    x->x_h = i;
    my_canvas_draw(x, x->x_glist, 0);
}

static void my_canvas_color(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);

    x->x_bcol = (int)atom_getintarg(0, ac, av)&0xffffff;
    x->x_lcol = (int)atom_getintarg(1, ac, av)&0xffffff;
    sys_vgui(".x%x.c itemconfigure %xRECT -fill #%6.6x -outline #%6.6x\n", canvas, x,
	     x->x_bcol, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x,
	         x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_bcol);
    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);
}

static void my_canvas_send(t_my_canvas *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void my_canvas_receive(t_my_canvas *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    x->x_rcv_able = rcvable;
}

static void my_canvas_label(t_my_canvas *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    x->x_lab = s;
    if(!strcmp(s->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		 glist_getcanvas(x->x_glist), x, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n",
		 glist_getcanvas(x->x_glist), x, s->s_name);
}

static void my_canvas_label_pos(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    sys_vgui(".x%x.c coords %xLABEL %d %d\n", glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
}

static void my_canvas_label_font(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
}

static void my_canvas_list(t_my_canvas *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	t_symbol *key = atom_getsymbolarg(1, ac, av);
	int keydown = atom_getintarg(0, ac, av);

	if(keydown)
	{
	    int refresh = 1,i,d=1;
	    static char buf[20];

	    buf[0] = 0;
	    if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		x->x_shiftdown = 1;
	    else
	    {
		if(x->x_shiftdown)
		    d = 10;

		if(!strcmp(key->s_name, "Up"))
		    x->x_obj.te_ypos -= d;
		else if(!strcmp(key->s_name, "Down"))
		    x->x_obj.te_ypos += d;
		else if(!strcmp(key->s_name, "Left"))
		    x->x_obj.te_xpos -= d;
		else if(!strcmp(key->s_name, "Right"))
		    x->x_obj.te_xpos += d;
		else
		    refresh = 0;
		if(refresh)
		{
		    my_canvas_draw(x, x->x_glist, 0);
		}
	    }
	}
	else
	{
	    if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		x->x_shiftdown = 0;
	}
    }
    else
        post("error: my_canvas: no method for 'list'");
}

static void *my_canvas_new(t_symbol *s, int argc, t_atom *argv)
{
    t_my_canvas *x = (t_my_canvas *)pd_new(my_canvas_class);
    int bcol= -233017;
    int lcol= -66577;
    int a=IEM_GUI_DEFAULTSIZE, w=100, h=60;
    int ldx=20, ldy=12, f=2, fs=20;
    int sndable=1, rcvable=1, i=0;
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc >= 10)&&(argc <= 12))
       &&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)&&IS_A_FLOAT(argv,2))
    {
	a = (int)atom_getintarg(0, argc, argv);
	w = (int)atom_getintarg(1, argc, argv);
	h = (int)atom_getintarg(2, argc, argv);
    }
    if((argc == 12)&&IS_A_SYMBOL(argv,3)&&IS_A_SYMBOL(argv,4))
    {
	i = 2;
	snd = atom_getsymbolarg(3, argc, argv);
	rcv = atom_getsymbolarg(4, argc, argv);
    }
    else if((argc == 11)&&IS_A_SYMBOL(argv,3))
    {
	i = 1;
	rcv = atom_getsymbolarg(3, argc, argv);
    }

    if(((argc >= 10)&&(argc <= 12))
       &&IS_A_SYMBOL(argv,i+3)&&IS_A_FLOAT(argv,i+4)
       &&IS_A_FLOAT(argv,i+5)&&IS_A_FLOAT(argv,i+6)
       &&IS_A_FLOAT(argv,i+7)&&IS_A_FLOAT(argv,i+8)
       &&IS_A_FLOAT(argv,i+9))
    {
	lab = atom_getsymbolarg(i+3, argc, argv);
	ldx = (int)atom_getintarg(i+4, argc, argv);
	ldy = (int)atom_getintarg(i+5, argc, argv);
	f = (int)atom_getintarg(i+6, argc, argv);
	fs = (int)atom_getintarg(i+7, argc, argv);
	bcol = (int)atom_getintarg(i+8, argc, argv);
	lcol = (int)atom_getintarg(i+9, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(a < 1)
	a = 1;
    x->x_size = a;
    if(w < 1)
	w = 1;
    x->x_w = w;
    if(h < 1)
	h = 1;
    x->x_h = h;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
	fs = 4;
    x->x_fontsize = fs;
    if(lcol < 0)
    {
	lcol = -1 - lcol;
	x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
	bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    x->x_shiftdown = 0;
    x->x_at[0].a_type = A_FLOAT;
    x->x_at[1].a_type = A_FLOAT;
    return (x);
}

static void my_canvas_ff(t_my_canvas *x)
{
    if(x->x_selected)
	pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void my_canvas_setup(void)
{
    my_canvas_class = class_new(gensym("cnv"), (t_newmethod)my_canvas_new,
			     (t_method)my_canvas_ff, sizeof(t_my_canvas), CLASS_NOINLET, A_GIMME, 0);
    class_addcreator((t_newmethod)my_canvas_new, gensym("my_canvas"), A_GIMME, 0);
    class_addlist(my_canvas_class, my_canvas_list);
    class_addmethod(my_canvas_class, (t_method)my_canvas_dialog, gensym("dialog"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_size, gensym("size"), A_DEFFLOAT, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_vis_size, gensym("vis_size"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_color, gensym("color"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(my_canvas_class, (t_method)my_canvas_get_pos, gensym("get_pos"), 0);

    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    my_canvas_widgetbehavior.w_getrectfn = my_canvas_getrect;
    my_canvas_widgetbehavior.w_displacefn = my_canvas_displace;
    my_canvas_widgetbehavior.w_selectfn = my_canvas_select;
    my_canvas_widgetbehavior.w_activatefn = NULL;
    my_canvas_widgetbehavior.w_deletefn = my_canvas_delete;
    my_canvas_widgetbehavior.w_visfn = my_canvas_vis;
#if (PD_VERSION_MINOR > 31)
    my_canvas_widgetbehavior.w_clickfn = NULL;
    my_canvas_widgetbehavior.w_propertiesfn = my_canvas_properties;
#endif
    my_canvas_widgetbehavior.w_savefn = my_canvas_save;
    class_setwidget(my_canvas_class, &my_canvas_widgetbehavior);
    class_sethelpsymbol(my_canvas_class, gensym("my_canvas"));
}
