/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution. */

/* iem-gui-lib written by Thomas Musil (c) IEM KUG Graz Austria 2000-2001 */

/* ------------ vertical  slider ----------------------- */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "m_imp.h"
#include "g_canvas.h"
#include "t_tk.h"
#include "g_7_guis.h"

typedef struct _vslider
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_w;
     int      x_h;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_pos;
     int      x_val;
     int      x_lcol;
     int      x_fcol;
     int      x_bcol;
     int      x_lin0_log1;
     int      x_loadinit;
     int      x_steady;
     int      x_snd_able;
     int      x_rcv_able;
     int      x_finemoved;
     int      x_selected;
     int      x_shiftdown;
     double   x_min;
     double   x_max;
     double   x_k;
     t_symbol *x_snd;
     t_symbol *x_rcv;
     t_symbol *x_lab;
} t_vslider;

t_widgetbehavior vslider_widgetbehavior;
static t_class *vslider_class;

/* widget helper functions */

static void vslider_update(t_vslider *x, t_glist *glist)
{
    if (glist_isvisible(glist))
    {
	int r = x->x_obj.te_ypos + x->x_h - (x->x_val + 50)/100;
	sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 glist_getcanvas(glist), x, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r);
    }
}

static void vslider_draw(t_vslider *x, t_glist *glist, int firsttime)
{
    int r = x->x_obj.te_ypos + x->x_h - (x->x_val + 50)/100;
    t_canvas *canvas=glist_getcanvas(glist);

    if (firsttime)
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h+3,
		 x->x_bcol, x);
	sys_vgui(".x%x.c create line %d %d %d %d -width 3 -fill #%6.6x -tags %xKNOB\n",
		 canvas, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r, x->x_fcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos, x->x_obj.te_ypos + x->x_h+2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos + x->x_h+3,
		 x, 0);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos-1,
		 x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + x->x_w, x->x_obj.te_ypos + x->x_h+3);
        sys_vgui(".x%x.c coords %xKNOB %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos+1, r,
		 x->x_obj.te_xpos + x->x_w, r);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos, x->x_obj.te_ypos + x->x_h+2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos + x->x_h+3);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+7, x->x_obj.te_ypos-1);
    }
}

static void vslider_erase(t_vslider* x,t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xKNOB\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
}

/* ------------------------ hsld widgetbehaviour----------------------------- */


static void vslider_getrect(t_gobj *z, t_glist *owner,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_vslider* x = (t_vslider*)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos - 2;
    *xp2 = x->x_obj.te_xpos + x->x_w;
    *yp2 = x->x_obj.te_ypos + x->x_h + 3;
}

static void vslider_displace(t_gobj *z, t_glist *glist,
    int dx, int dy)
{
    t_vslider* x = (t_vslider*)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    vslider_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text*)z);
}

static void vslider_select(t_gobj *z, t_glist *glist, int selected)
{
    t_vslider *x = (t_vslider *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void vslider_delete(t_gobj *z, t_glist *glist)
{
    t_vslider *x = (t_vslider *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void vslider_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_vslider *x = (t_vslider *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 vslider_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 vslider_erase(x, glist);
         rtext_free(y);
    }
}

static void vslider_save(t_gobj *z, t_binbuf *b)
{
    t_vslider *x = (t_vslider *)z;
    int bcol, fcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    fcol = -1 - (((0xfc0000 & x->x_fcol) >> 6)|((0xfc00 & x->x_fcol) >> 4)|((0xfc & x->x_fcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiiffiisssiiiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("vslider"), x->x_w, x->x_h,
		(float)x->x_min, (float)x->x_max,
                x->x_lin0_log1, x->x_loadinit,
		x->x_snd, x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8, x->x_fontsize,
		bcol, fcol, lcol,
	        x->x_val, x->x_steady);
    binbuf_addv(b, ";");
}

void vslider_check_height(t_vslider *x, int h)
{
    if(h < IEM_SL_MINSIZE)
	h = IEM_SL_MINSIZE;
    x->x_h = h;
    if(x->x_val > (x->x_h*100 - 100))
    {
	x->x_pos = x->x_h*100 - 100;
	x->x_val = x->x_pos;
    }
    if(x->x_lin0_log1)
	x->x_k = log(x->x_max/x->x_min)/(double)(x->x_h - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

void vslider_check_minmax(t_vslider *x, double min, double max)
{
    if(x->x_lin0_log1)
    {
	if((min == 0.0)&&(max == 0.0))
            max = 1.0;
	if(max > 0.0)
	{
	    if(min <= 0.0)
                min = 0.01*max;
	}
	else
	{
            if(min > 0.0)
                max = 0.01*min;
	}
    }
    x->x_min = min;
    x->x_max = max;
    if(x->x_lin0_log1)
        x->x_k = log(x->x_max/x->x_min)/(double)(x->x_h - 1);
    else
	x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

static void vslider_properties(t_gobj *z, t_glist *owner)
{
    t_vslider *x = (t_vslider *)z;
    char buf[600];
    int steady=-23;

    if(x->x_steady)
        steady = -22;
    str_xch_dollar2raute(x->x_snd);
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s VSLIDER \
	    --------dimensions(pix)(pix):-------- %g %g width: %g %g height: \
	    -----------output-range:----------- %g bottom: %g top: %g \
	    %g lin log %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_w, (float)IEM_GUI_MINSIZE, (float)x->x_h, (float)IEM_SL_MINSIZE,
	    x->x_min, x->x_max, 0.0,/*no_schedule*/
	    (float)x->x_lin0_log1, (float)x->x_loadinit, (float)steady,/*no multi, but iem-characteristic*/
	    x->x_snd->s_name, x->x_rcv->s_name,
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, 0xffffff & x->x_fcol, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_snd);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void vslider_bang(t_vslider *x)
{
    double out;

    if(x->x_lin0_log1)
	out = x->x_min*exp(x->x_k*(double)(x->x_val)*0.01);
    else
	out = (double)(x->x_val)*0.01*x->x_k + x->x_min;
    if((out < 1.0e-10)&&(out > -1.0e-10))
        out = 0.0;

    outlet_float(x->x_obj.ob_outlet, out);
    if(x->x_snd_able && x->x_snd->s_thing)
	pd_float(x->x_snd->s_thing, out);
}

static void vslider_dialog(t_vslider *x, t_symbol *s, int argc, t_atom *argv)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int w = (int)atom_getintarg(0, argc, argv);
    int h = (int)atom_getintarg(1, argc, argv);
    double min = (double)atom_getfloatarg(2, argc, argv);
    double max = (double)atom_getfloatarg(3, argc, argv);
    int lilo = (int)atom_getintarg(4, argc, argv);
    int init = (int)atom_getintarg(5, argc, argv);
    int steady = (int)atom_getintarg(6, argc, argv);
    t_symbol *snd = atom_getsymbolarg(7, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int fcol = (int)atom_getintarg(15, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int sndable=1, rcvable=1;

    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(init != 0) init = 1;
    x->x_loadinit = init;
    if(steady == 3)
	x->x_steady = 0;
    else
        x->x_steady = 1;
    str_xch_raute2dollar(snd);
    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }

    x->x_snd = snd;
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_w = iem_gui_clip_size(w);
    vslider_check_height(x, h);
    vslider_check_minmax(x, min, max);
    x->x_lcol = lcol & 0xffffff;
    x->x_fcol = fcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;

    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);
    sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas,
	     x, x->x_fcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
	     x, x->x_bcol);
    vslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(canvas, (t_text*)x);
}

static void vslider_motion(t_vslider *x, t_floatarg dx, t_floatarg dy)
{
    int old = x->x_val;

    if(x->x_finemoved)
        x->x_pos -= (int)dy;
    else
	x->x_pos -= 100*(int)dy;
    x->x_val = x->x_pos;
    if(x->x_val > (100*x->x_h - 100))
    {
	x->x_val = 100*x->x_h - 100;
        x->x_pos += 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(x->x_val < 0)
    {
	x->x_val = 0;
	x->x_pos -= 50;
	/*x->x_pos /= 100;
	 x->x_pos *= 100;*/
        x->x_pos -= x->x_pos%100;
    }
    if(old != x->x_val)
    {
	vslider_update(x, x->x_glist);
	vslider_bang(x);
    }
}

static void vslider_click(t_vslider *x, t_floatarg xpos, t_floatarg ypos,
		       t_floatarg shift, t_floatarg ctrl, t_floatarg alt)
{
    if(!x->x_steady)
	x->x_val = (int)(100.0 * (x->x_h + x->x_obj.te_ypos - ypos));
    if(x->x_val > (100*x->x_h - 100))
	x->x_val = 100*x->x_h - 100;
    if(x->x_val < 0)
	x->x_val = 0;
    x->x_pos = x->x_val;
    vslider_update(x, x->x_glist);
    vslider_bang(x);
#if (PD_VERSION_MINOR > 31)
    glist_grab(x->x_glist, &x->x_obj.te_g, (t_glistmotionfn)vslider_motion,
	       0, xpos, ypos);
#else
    glist_grab(x->x_glist, &x->x_obj.te_g, xpos, ypos);
#endif
}

#if (PD_VERSION_MINOR > 31)
static int vslider_newclick(t_gobj *z, struct _glist *glist,
    int xpix, int ypix, int shift, int alt, int dbl, int doit)
{
    t_vslider* x = (t_vslider *)z;

    if(doit)
    {
	vslider_click( x, (t_floatarg)xpix, (t_floatarg)ypix, (t_floatarg)shift,
		    0, (t_floatarg)alt);
	if(shift)
	    x->x_finemoved = 1;
	else
            x->x_finemoved = 0;
    }
    return (1);
}
#endif

static void vslider_set(t_vslider *x, t_floatarg f)
{
    double g;

    if(f > x->x_max)
	f = x->x_max;
    if(f < x->x_min)
	f = x->x_min;
    if(x->x_lin0_log1)
	g = log(f/x->x_min)/x->x_k;
    else
	g = (f - x->x_min) / x->x_k;
    x->x_val = (int)(100.0*g + 0.49999);
    x->x_pos = x->x_val;
    vslider_update(x, x->x_glist);
}

static void vslider_size(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_w = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    vslider_check_height(x, (int)atom_getintarg(1, ac, av));
    vslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vslider_delta(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    vslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vslider_pos(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    vslider_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vslider_range(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    vslider_check_minmax(x, (double)atom_getfloatarg(0, ac, av),
			 (double)atom_getfloatarg(1, ac, av));
}

static void vslider_color(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);

    x->x_bcol = (int)atom_getintarg(0, ac, av)&0xffffff;
    x->x_fcol = (int)atom_getintarg(1, ac, av)&0xffffff;
    x->x_lcol = (int)atom_getintarg(2, ac, av)&0xffffff;
    sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas,
	     x, x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
    sys_vgui(".x%x.c itemconfigure %xKNOB -fill #%6.6x\n", canvas,
	     x, x->x_fcol);
    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas,
	     x, x->x_bcol);
}

static void vslider_send(t_vslider *x, t_symbol *s)
{
    int sndable=1;

    str_xch_raute2dollar(s);
    if(!strcmp(s->s_name, "empty")) sndable = 0;
    x->x_snd = s;
    x->x_snd_able = sndable;
}

static void vslider_receive(t_vslider *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    x->x_rcv_able = rcvable;
}

static void vslider_label(t_vslider *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    x->x_lab = s;
    if(!strcmp(s->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n", glist_getcanvas(x->x_glist), x, "");
    else
        sys_vgui(".x%x.c itemconfigure %xLABEL -text {%s} \n", glist_getcanvas(x->x_glist), x, s->s_name);
}

static void vslider_label_pos(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    sys_vgui(".x%x.c coords %xLABEL %d %d\n", glist_getcanvas(x->x_glist), x,
	     x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
}

static void vslider_label_font(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    int f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold}\n",
	     glist_getcanvas(x->x_glist), x, x->x_font, x->x_fontsize);
}

static void vslider_log(t_vslider *x)
{
    x->x_lin0_log1 = 1;
    vslider_check_minmax(x, x->x_min, x->x_max);
}

static void vslider_lin(t_vslider *x)
{
    x->x_lin0_log1 = 0;
    x->x_k = (x->x_max - x->x_min)/(double)(x->x_h - 1);
}

static void vslider_init(t_vslider *x, t_floatarg f)
{
    x->x_loadinit = (f==0.0)?0:1;
}

static void vslider_steady(t_vslider *x, t_floatarg f)
{
    x->x_steady = (f==0.0)?0:1;
}

static void vslider_float(t_vslider *x, t_floatarg f)
{
    vslider_set(x, f);
    vslider_bang(x);
}

static void vslider_loadbang(t_vslider *x)
{
    if(!sys_noloadbang && x->x_loadinit)
    {
	vslider_update(x, x->x_glist);
	vslider_bang(x);
    }
}

static void vslider_list(t_vslider *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto vsliderlistelse;
	else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;
		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
                    {
			vslider_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
	return;
    }
vsliderlistelse:
    if(IS_A_FLOAT(av,0))
	vslider_float(x, atom_getfloatarg(0, ac, av));
}

static void *vslider_new(t_symbol *s, int argc, t_atom *argv)
{
    t_vslider *x = (t_vslider *)pd_new(vslider_class);
    int bcol= -262144;
    int fcol= -1;
    int lcol= -1;
    int w=IEM_GUI_DEFAULTSIZE, h=IEM_SL_DEFAULTSIZE;
    int lilo=0, init=0, ldx=20, ldy=8, f=0, fs=10;
    int sndable=1, rcvable=1, v=0, steady=1;
    double min=0.0, max=(double)(IEM_SL_DEFAULTSIZE-1);
    t_symbol *snd = gensym("empty"), *rcv = gensym("empty"), *lab = gensym("empty");

    if(((argc == 17)||(argc == 18))&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)
       &&IS_A_FLOAT(argv,4)&&IS_A_FLOAT(argv,5)
       &&IS_A_SYMBOL(argv,6)&&IS_A_SYMBOL(argv,7)
       &&IS_A_SYMBOL(argv,8)&&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10)
       &&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)
       &&IS_A_FLOAT(argv,14)&&IS_A_FLOAT(argv,15)&&IS_A_FLOAT(argv,16))
    {
        w = (int)atom_getintarg(0, argc, argv);
	h = (int)atom_getintarg(1, argc, argv);
	min = (double)atom_getfloatarg(2, argc, argv);
	max = (double)atom_getfloatarg(3, argc, argv);
	lilo = (int)atom_getintarg(4, argc, argv);
	init = (int)atom_getintarg(5, argc, argv);
        snd = atom_getsymbolarg(6, argc, argv);
	rcv = atom_getsymbolarg(7, argc, argv);
	lab = atom_getsymbolarg(8, argc, argv);
	ldx = (int)atom_getintarg(9, argc, argv);
	ldy = (int)atom_getintarg(10, argc, argv);
        f = (int)atom_getintarg(11, argc, argv);
        fs = (int)atom_getintarg(12, argc, argv);
        bcol = (int)atom_getintarg(13, argc, argv);
	fcol = (int)atom_getintarg(14, argc, argv);
        lcol = (int)atom_getintarg(15, argc, argv);
	v = (int)atom_getintarg(16, argc, argv);
    }
    if((argc == 18)&&IS_A_FLOAT(argv,17))
	steady = (int)atom_getintarg(17, argc, argv);
    if(init != 0) init = 1;
    x->x_loadinit = init;
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(x->x_loadinit)
	x->x_val = v;
    else
        x->x_val = 0;
    x->x_pos = x->x_val;
    if(lilo != 0) lilo = 1;
    x->x_lin0_log1 = lilo;
    if(steady != 0) steady = 1;
    x->x_steady = steady;
    if(!strcmp(snd->s_name, "empty")) sndable = 0;
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_snd_able = sndable;
    x->x_rcv_able = rcvable;
    x->x_snd = snd;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_w = iem_gui_clip_size(w);
    vslider_check_height(x, h);
    vslider_check_minmax(x, min, max);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(fcol < 0)
    {
        fcol = -1 - fcol;
	x->x_fcol = ((fcol & 0x3f000) << 6)|((fcol & 0xfc0) << 4)|((fcol & 0x3f) << 2);
    }
    else
    {
	fcol = iem_gui_modulo_color(fcol);
	x->x_fcol = iem_gui_color_hex[fcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_finemoved = 0;
    x->x_selected = 0;
    x->x_shiftdown = 0;
    outlet_new(&x->x_obj, &s_float);
    return (x);
}

static void vslider_free(t_vslider *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void vslider_setup(void)
{
    vslider_class = class_new(gensym("vsl"), (t_newmethod)vslider_new,
			      (t_method)vslider_free, sizeof(t_vslider), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)vslider_new, gensym("vslider"), A_GIMME, 0);
    class_addbang(vslider_class,vslider_bang);
    class_addfloat(vslider_class,vslider_float);
    class_addlist(vslider_class, vslider_list);
    class_addmethod(vslider_class, (t_method)vslider_click, gensym("click"),
		    A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_motion, gensym("motion"),
		    A_FLOAT, A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_loadbang, gensym("loadbang"), 0);
    class_addmethod(vslider_class, (t_method)vslider_set, gensym("set"), A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_size, gensym("size"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_range, gensym("range"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_color, gensym("color"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_send, gensym("send"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(vslider_class, (t_method)vslider_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_label_font, gensym("label_font"), A_GIMME, 0);
    class_addmethod(vslider_class, (t_method)vslider_log, gensym("log"), 0);
    class_addmethod(vslider_class, (t_method)vslider_lin, gensym("lin"), 0);
    class_addmethod(vslider_class, (t_method)vslider_init, gensym("init"), A_FLOAT, 0);
    class_addmethod(vslider_class, (t_method)vslider_steady, gensym("steady"), A_FLOAT, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    vslider_widgetbehavior.w_getrectfn =    vslider_getrect;
    vslider_widgetbehavior.w_displacefn =   vslider_displace;
    vslider_widgetbehavior.w_selectfn =     vslider_select;
    vslider_widgetbehavior.w_activatefn =   NULL;
    vslider_widgetbehavior.w_deletefn =     vslider_delete;
    vslider_widgetbehavior.w_visfn =        vslider_vis;
#if (PD_VERSION_MINOR > 31)
    vslider_widgetbehavior.w_clickfn =      vslider_newclick;
    vslider_widgetbehavior.w_propertiesfn = vslider_properties;;
#endif
    vslider_widgetbehavior.w_savefn =       vslider_save;
    class_setwidget(vslider_class, &vslider_widgetbehavior);
    class_sethelpsymbol(vslider_class, gensym("vslider"));
}
