/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution. */

/* iem-gui-lib written by Thomas Musil (c) IEM KUG Graz Austria 2000-2001 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "m_imp.h"
#include "g_canvas.h"
#include "t_tk.h"
#include "g_7_guis.h"

/* ----- peak- & rms- vu-meter-display ---------- */

int iem_gui_vu_db2i[]=
{
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
     9, 9, 9, 9, 9,10,10,10,10,10,
    11,11,11,11,11,12,12,12,12,12,
    13,13,13,13,14,14,14,14,15,15,
    15,15,16,16,16,16,17,17,17,18,
    18,18,19,19,19,20,20,20,21,21,
    22,22,23,23,24,24,25,26,27,28,
    29,30,31,32,33,33,34,34,35,35,
    36,36,37,37,37,38,38,38,39,39,
    39,39,39,39,40,40
};

int iem_gui_vu_col[]=
{
    0,17,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
    15,15,15,15,15,15,15,15,15,15,14,14,13,13,13,13,13,13,13,13,13,13,13,19,19,19
};

char *iem_gui_vu_scale_str[]=
{
    "",
    "<-99",
    "",
    "",
    "",
    "-50",
    "",
    "",
    "",
    "-30",
    "",
    "",
    "",
    "-20",
    "",
    "",
    "",
    "-12",
    "",
    "",
    "",
    "-6",
    "",
    "",
    "",
    "-2",
    "",
    "",
    "",
    "-0dB",
    "",
    "",
    "",
    "+2",
    "",
    "",
    "",
    "+6",
    "",
    "",
    "",
    ">+12",
    "",
    "",
    "",
    "",
    "",
};

typedef struct _vu
{
     t_object x_obj;
     t_glist *x_glist;
     int      x_w;
     int      x_h;
     int      x_led_size;
     int      x_ldx;
     int      x_ldy;
     char     x_font[16];
     int      x_fontsize;
     int      x_peak;
     int      x_rms;
     int      x_lcol;
     int      x_bcol;
     int      x_rcv_able;
     int      x_scale;
     int      x_selected;
     int      x_shiftdown;
     t_symbol *x_rcv;
     t_symbol *x_lab;
     void     *x_out_rms;
     void     *x_out_peak;
} t_vu;

t_widgetbehavior vu_widgetbehavior;
static t_class *vu_class;

/* widget helper functions */

static void vu_update_rms(t_vu *x, t_glist *glist)
{
    if(glist_isvisible(glist))
    {
	int w4=x->x_w/4, off=x->x_obj.te_ypos-1;
	int quad1=x->x_obj.te_xpos+w4+1, quad3=x->x_obj.te_xpos+x->x_w-w4-1;

	sys_vgui(".x%x.c coords %xRCOVER %d %d %d %d\n",
		 glist_getcanvas(glist), x, quad1, off, quad3,
		 off + (x->x_led_size+1)*(IEM_VU_STEPS-x->x_rms));
    }
}

static void vu_update_peak(t_vu *x, t_glist *glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(glist_isvisible(glist))
    {
	if(x->x_peak)
	{
	    int i=iem_gui_vu_col[x->x_peak];
	    int j=x->x_obj.te_ypos + (x->x_led_size+1)*(IEM_VU_STEPS+1-x->x_peak)
		- (x->x_led_size+1)/2;

            sys_vgui(".x%x.c coords %xPLED %d %d %d %d\n", canvas, x,
		     x->x_obj.te_xpos, j,
		     x->x_obj.te_xpos+x->x_w+1, j);
	    sys_vgui(".x%x.c itemconfigure %xPLED -fill #%6.6x\n", canvas, x,
		     iem_gui_color_hex[i]);
	}
	else
	{
	    int mid=x->x_obj.te_xpos+x->x_w/2;

            sys_vgui(".x%x.c itemconfigure %xPLED -fill #%6.6x\n",
		     canvas, x, x->x_bcol);
            sys_vgui(".x%x.c coords %xPLED %d %d %d %d\n",
		 canvas, x, mid, x->x_obj.te_ypos+20,
		     mid, x->x_obj.te_ypos+20);
	}
    }
}

static void vu_draw(t_vu *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if (firsttime)
    {
	int w4=x->x_w/4, mid=x->x_obj.te_xpos+x->x_w/2, quad1=x->x_obj.te_xpos+w4+1;
	int quad3=x->x_obj.te_xpos+x->x_w-w4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int led_col, yyy, i, k4=x->x_obj.te_ypos-k3;

	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1,x->x_obj.te_ypos+x->x_h+2, x->x_bcol, x);
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    led_col = iem_gui_vu_col[i];
            yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xRLED%d\n",
		     canvas, quad1, yyy, quad3, yyy, x->x_led_size, iem_gui_color_hex[led_col], x, i);
	    if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
			 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
			 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
			 x->x_lcol, x, i);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
		     canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		     x->x_lcol, x, i);
	}
        sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -outline #%6.6x -tags %xRCOVER\n",
		 canvas, quad1, x->x_obj.te_ypos-1, quad3-1,
		 x->x_obj.te_ypos-1 + k1*IEM_VU_STEPS, x->x_bcol, x->x_bcol, x);
        sys_vgui(".x%x.c create line %d %d %d %d -width %d -fill #%6.6x -tags %xPLED\n",
		 canvas, mid, x->x_obj.te_ypos+10,
		     mid, x->x_obj.te_ypos+10, x->x_led_size, x->x_bcol, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize, x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize, x->x_lcol, x);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos + x->x_h+2,
		 x, 0);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos + x->x_h+2,
		 x, 1);
	sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos-1,
		 x, 0);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos-1,
		 x, 1);
    }
    else
    {
	int w4=x->x_w/4, quad1=x->x_obj.te_xpos+w4+1;
	int quad3=x->x_obj.te_xpos+x->x_w-w4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int yyy, i, k4=x->x_obj.te_ypos-k3;

        sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1,x->x_obj.te_ypos+x->x_h+2);
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
            yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c coords %xRLED%d %d %d %d %d\n",
		     canvas, x, i, quad1, yyy, quad3, yyy);
	    if(((i+2)&3) && (x->x_scale))
                sys_vgui(".x%x.c coords %xSCALE%d %d %d\n",
		     canvas, x, i, end, yyy+k3);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    yyy = k4 + k1*(k2-i);
	    sys_vgui(".x%x.c coords %xSCALE%d %d %d\n",
		 canvas, x, i, end, yyy+k3);
	}
        vu_update_peak(x, glist);
	vu_update_rms(x, glist);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx,
		 x->x_obj.te_ypos+x->x_ldy);
	sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos + x->x_h+2);
        sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 1,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos + x->x_h+1,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos + x->x_h+2);
	sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0,
		 x->x_obj.te_xpos-1, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos + IOWIDTH-1, x->x_obj.te_ypos-1);
        sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 1,
		 x->x_obj.te_xpos+x->x_w+1-IOWIDTH, x->x_obj.te_ypos-2,
		 x->x_obj.te_xpos+x->x_w+1, x->x_obj.te_ypos-1);
    }
}

static void vu_confitem(t_vu* x,t_glist* glist)
{
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x, x->x_bcol);
    for(i=1; i<=IEM_VU_STEPS; i++)
    {
        sys_vgui(".x%x.c itemconfigure %xRLED%d -width %d\n", canvas, x, i,
		     x->x_led_size);
	if(((i+2)&3) && (x->x_scale))
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -text {%s} -font {%s %d bold} -fill #%6.6x\n",
		     canvas, x, i, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
    }
    if(x->x_scale)
    {
	i=IEM_VU_STEPS+1;
	sys_vgui(".x%x.c itemconfigure %xSCALE%d -text {%s} -font {%s %d bold} -fill #%6.6x\n",
		     canvas, x, i, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol);
    }
    if(!strcmp(x->x_lab->s_name, "empty"))
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
    else
	sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		 canvas, x, x->x_font, x->x_fontsize,
		 x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		 x->x_lab->s_name);

    sys_vgui(".x%x.c itemconfigure %xRCOVER -fill #%6.6x -outline #%6.6x\n", canvas,
	     x, x->x_bcol, x->x_bcol);
    sys_vgui(".x%x.c itemconfigure %xPLED -width %d\n", canvas, x,
		     x->x_led_size);
}

static void vu_erase(t_vu* x,t_glist* glist)
{
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    for(i=1; i<=IEM_VU_STEPS; i++)
    {
	sys_vgui(".x%x.c delete %xRLED%d\n", canvas, x, i);
        if(((i+2)&3) && (x->x_scale))
            sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
    }
    if(x->x_scale)
    {
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
    }
    sys_vgui(".x%x.c delete %xPLED\n", canvas, x);
    sys_vgui(".x%x.c delete %xRCOVER\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 1);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 1);
}

/* ------------------------ vu widgetbehaviour----------------------------- */


static void vu_getrect(t_gobj *z, t_glist *owner,
    int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_vu* x = (t_vu*)z;

    *xp1 = x->x_obj.te_xpos - 1;
    *yp1 = x->x_obj.te_ypos - 2;
    *xp2 = x->x_obj.te_xpos + x->x_w + 1;
    *yp2 = x->x_obj.te_ypos + x->x_h + 2;
}

static void vu_displace(t_gobj *z, t_glist *glist,
    int dx, int dy)
{
    t_vu* x = (t_vu*)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    vu_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void vu_select(t_gobj *z, t_glist *glist, int selected)
{
    t_vu *x = (t_vu *)z;
    int i;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
			 canvas, x, i, IEM_GUI_COLOR_SELECTED);
	}
	if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
		 canvas, x, i, IEM_GUI_COLOR_SELECTED);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        for(i=1; i<=IEM_VU_STEPS; i++)
	{
            if(((i+2)&3) && (x->x_scale))
		sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
			 canvas, x, i, x->x_lcol);
	}
        if(x->x_scale)
	{
	    i=IEM_VU_STEPS+1;
	    sys_vgui(".x%x.c itemconfigure %xSCALE%d -fill #%6.6x\n",
		  canvas, x, i, x->x_lcol);
	}
	sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void vu_delete(t_gobj *z, t_glist *glist)
{
    t_vu *x = (t_vu *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void vu_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_vu *x = (t_vu *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 vu_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 vu_erase(x, glist);
         rtext_free(y);
    }
}

static void vu_save(t_gobj *z, t_binbuf *b)
{
    t_vu *x = (t_vu *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisiissiiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		gensym("vu"), x->x_w, x->x_h,
                x->x_rcv, x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8, x->x_fontsize,
		bcol, lcol, x->x_scale);
    binbuf_addv(b, ";");
}

void vu_check_height(t_vu *x, int h)
{
    int n;

    n = h / IEM_VU_STEPS;
    if(n < IEM_VU_MINSIZE)
	n = IEM_VU_MINSIZE;
    x->x_led_size = n-1;
    x->x_h = IEM_VU_STEPS * n;
}

static void vu_scale(t_vu *x, t_floatarg fscale)
{
    int i, scale = (int)fscale;

    if(scale != 0) scale = 1;
    if(x->x_scale && !scale)
    {
	t_canvas *canvas=glist_getcanvas(x->x_glist);

	x->x_scale = (int)scale;
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    if((i+2)&3)
		sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
	}
	i=IEM_VU_STEPS+1;
	sys_vgui(".x%x.c delete %xSCALE%d\n", canvas, x, i);
    }
    if(!x->x_scale && scale)
    {
	int w4=x->x_w/4, end=x->x_obj.te_xpos+x->x_w+4;
	int k1=x->x_led_size+1, k2=IEM_VU_STEPS+1, k3=k1/2;
	int yyy, k4=x->x_obj.te_ypos-k3;
	t_canvas *canvas=glist_getcanvas(x->x_glist);

	x->x_scale = (int)scale;
	for(i=1; i<=IEM_VU_STEPS; i++)
	{
	    yyy = k4 + k1*(k2-i);
	    if((i+2)&3)
		sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
			 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
			 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
			 x->x_lcol, x, i);
	}
	i=IEM_VU_STEPS+1;
	yyy = k4 + k1*(k2-i);
	sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		 -font {%s %d bold} -fill #%6.6x -tags %xSCALE%d\n",
		 canvas, end, yyy+k3, iem_gui_vu_scale_str[i], x->x_font, x->x_fontsize,
		 x->x_lcol, x, i);
    }
}

static void vu_properties(t_gobj *z, t_glist *owner)
{
    t_vu *x = (t_vu *)z;
    char buf[600];
    
    str_xch_dollar2raute(x->x_rcv);
    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s VU-METER \
	    --------dimensions(pix)(pix):-------- %g %g width: %g %g height: \
	    empty 0.0 empty 0.0 empty %g \
	    %g no_scale scale %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
	    (float)x->x_w, (float)IEM_GUI_MINSIZE, (float)x->x_h, (float)(IEM_VU_STEPS*IEM_VU_MINSIZE),
	    0.0,/*no_schedule*/
	    (float)x->x_scale, -1.0, -1.0,/*no linlog, no init, no multi*/
	    "nosndno", x->x_rcv->s_name,/*no send*/
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no front-color*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_rcv);
    str_xch_raute2dollar(x->x_lab);
}

static void vu_dialog(t_vu *x, t_symbol *s, int argc, t_atom *argv)
{
    int w = (int)atom_getintarg(0, argc, argv);
    int h = (int)atom_getintarg(1, argc, argv);
    int scale = (int)atom_getintarg(4, argc, argv);
    t_symbol *rcv = atom_getsymbolarg(8, argc, argv);
    t_symbol *lab = atom_getsymbolarg(9, argc, argv);
    int ldx = (int)atom_getintarg(10, argc, argv);
    int ldy = (int)atom_getintarg(11, argc, argv);
    int f = (int)atom_getintarg(12, argc, argv);
    int fs = (int)atom_getintarg(13, argc, argv);
    int bcol = (int)atom_getintarg(14, argc, argv);
    int lcol = (int)atom_getintarg(16, argc, argv);
    int rcvable=1, i;

    str_xch_raute2dollar(rcv);
    str_xch_raute2dollar(lab);
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(rcv->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = rcv;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    x->x_rcv_able = rcvable;
    x->x_w = iem_gui_clip_size(w);
    x->x_lcol = lcol & 0xffffff;
    x->x_bcol = bcol & 0xffffff;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    vu_check_height(x, h);
    if(scale != 0)
	scale = 1;
    vu_scale(x, (float)scale);
    vu_draw(x, x->x_glist, 0);
    vu_confitem(x, x->x_glist);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vu_size(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_w = iem_gui_clip_size((int)atom_getintarg(0, ac, av));
    vu_check_height(x, (int)atom_getintarg(1, ac, av));
    vu_draw(x, x->x_glist, 0);
    vu_confitem(x, x->x_glist);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vu_delta(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos += (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos += (int)atom_getintarg(1, ac, av);
    vu_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vu_pos(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_obj.te_xpos = (int)atom_getintarg(0, ac, av);
    x->x_obj.te_ypos = (int)atom_getintarg(1, ac, av);
    vu_draw(x, x->x_glist, 0);
    canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
}

static void vu_color(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_bcol = (int)atom_getintarg(0, ac, av)&0xffffff;
    x->x_lcol = (int)atom_getintarg(1, ac, av)&0xffffff;
    vu_confitem(x, x->x_glist);
}

static void vu_receive(t_vu *x, t_symbol *s)
{
    int rcvable=1;

    str_xch_raute2dollar(s);
    if(!strcmp(s->s_name, "empty")) rcvable = 0;
    if(rcvable)
    {
	if(strcmp(s->s_name, x->x_rcv->s_name))
	{
	    if(x->x_rcv_able)
		pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
            x->x_rcv = s;
            pd_bind(&x->x_obj.ob_pd, x->x_rcv);
	}
    }
    x->x_rcv_able = rcvable;
}

static void vu_label(t_vu *x, t_symbol *s)
{
    str_xch_raute2dollar(s);
    x->x_lab = s;
    vu_confitem(x, x->x_glist);
}

static void vu_label_pos(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    x->x_ldx = (int)atom_getintarg(0, ac, av);
    x->x_ldy = (int)atom_getintarg(1, ac, av);
    sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 glist_getcanvas(x->x_glist), x, x->x_obj.te_xpos+x->x_ldx,
	     x->x_obj.te_ypos+x->x_ldy);
}

static void vu_label_font(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    t_canvas *canvas=glist_getcanvas(x->x_glist);
    int i, f = (int)atom_getintarg(0, ac, av);

    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    f = (int)atom_getintarg(1, ac, av);
    if(f < 4)
        f = 4;
    x->x_fontsize = f;
    vu_confitem(x, x->x_glist);
}

static void vu_float(t_vu *x, t_floatarg rms)
{
    int i;

    if(rms <= IEM_VU_MINDB)
	x->x_rms = 0;
    else if(rms >= IEM_VU_MAXDB)
	x->x_rms = IEM_VU_STEPS;
    else
    {
	int i = (int)(2.0*(rms + IEM_VU_OFFSET));
	x->x_rms = iem_gui_vu_db2i[i];
    }
    i = (int)(100.0*rms + 10000.5);
    rms = 0.01*(float)(i - 10000);
    outlet_float(x->x_out_rms, rms);
    vu_update_rms(x, x->x_glist);
}

static void vu_ft1(t_vu *x, t_floatarg peak)
{
    int i;

    if(peak <= IEM_VU_MINDB)
	x->x_peak = 0;
    else if(peak >= IEM_VU_MAXDB)
	x->x_peak = IEM_VU_STEPS;
    else
    {
	int i = (int)(2.0*(peak + IEM_VU_OFFSET));
	x->x_peak = iem_gui_vu_db2i[i];
    }
    i = (int)(100.0*peak + 10000.5);
    peak = 0.01*(float)(i - 10000);
    outlet_float(x->x_out_peak, peak);
    vu_update_peak(x, x->x_glist);
}

static void vu_list(t_vu *x, t_symbol *s, int ac, t_atom *av)
{
    if(x->x_selected)
    {
	if((ac != 2)||!IS_A_FLOAT(av,0)||!IS_A_SYMBOL(av,1))
	    goto vulistelse;
        else
	{
	    t_symbol *key = atom_getsymbolarg(1, ac, av);
	    int keydown = atom_getintarg(0, ac, av);

	    if(keydown)
	    {
		int refresh = 1,i,d=1;
		static char buf[20];

		buf[0] = 0;
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 1;
		else
		{
		    if(x->x_shiftdown)
			d = 10;

		    if(!strcmp(key->s_name, "Up"))
			x->x_obj.te_ypos -= d;
		    else if(!strcmp(key->s_name, "Down"))
			x->x_obj.te_ypos += d;
		    else if(!strcmp(key->s_name, "Left"))
			x->x_obj.te_xpos -= d;
		    else if(!strcmp(key->s_name, "Right"))
			x->x_obj.te_xpos += d;
		    else
			refresh = 0;
		    if(refresh)
		    {
			vu_draw(x, x->x_glist, 0);
			canvas_fixlinesfor(glist_getcanvas(x->x_glist), (t_text*)x);
		    }
		}
	    }
	    else
	    {
		if(!strcmp(key->s_name, "Shift_L")||!strcmp(key->s_name, "Shift_R"))
		    x->x_shiftdown = 0;
	    }
	}
        return;
    }
vulistelse:
    if((IS_A_FLOAT(av,0))&&(IS_A_FLOAT(av,1)))
	{
	    vu_ft1(x, atom_getfloatarg(1, ac, av));
	    vu_float(x, atom_getfloatarg(0, ac, av));
	}
}

static void *vu_new(t_symbol *s, int argc, t_atom *argv)
{
    t_vu *x = (t_vu *)pd_new(vu_class);
    int bcol= -66577;
    int lcol= -1;
    int w=IEM_GUI_DEFAULTSIZE, h=IEM_VU_STEPS*IEM_VU_DEFAULTSIZE;
    int ldx=35, ldy=8, f=0, fs=10, scale=1;
    int rcvable=1;
    t_symbol *rcv = gensym("empty"), *lab = gensym("empty");

    if((argc == 11)&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_SYMBOL(argv,2)&&IS_A_SYMBOL(argv,3)
       &&IS_A_FLOAT(argv,4)&&IS_A_FLOAT(argv,5)
       &&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7)
       &&IS_A_FLOAT(argv,8)&&IS_A_FLOAT(argv,9)&&IS_A_FLOAT(argv,10))
    {
        w = (int)atom_getintarg(0, argc, argv);
	h = (int)atom_getintarg(1, argc, argv);
	rcv = atom_getsymbolarg(2, argc, argv);
	lab = atom_getsymbolarg(3, argc, argv);
	ldx = (int)atom_getintarg(4, argc, argv);
	ldy = (int)atom_getintarg(5, argc, argv);
        f = (int)atom_getintarg(6, argc, argv);
        fs = (int)atom_getintarg(7, argc, argv);
        bcol = (int)atom_getintarg(8, argc, argv);
	lcol = (int)atom_getintarg(9, argc, argv);
        scale = (int)atom_getintarg(10, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    if(!strcmp(rcv->s_name, "empty")) rcvable = 0;
    if(rcvable) pd_bind(&x->x_obj.ob_pd, rcv);
    x->x_rcv_able = rcvable;
    x->x_rcv = rcv;
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_w = iem_gui_clip_size(w);
    vu_check_height(x, h);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    if(scale != 0)
	scale = 1;
    x->x_scale = scale;
    x->x_selected = 0;
    x->x_shiftdown = 0;
    x->x_peak = 0;
    x->x_rms = 0;
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_float, gensym("ft1"));
    x->x_out_rms = outlet_new(&x->x_obj, &s_float);
    x->x_out_peak = outlet_new(&x->x_obj, &s_float);
    return (x);
}

static void vu_free(t_vu *x)
{
    if(x->x_selected)
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
    if(x->x_rcv_able)
	pd_unbind(&x->x_obj.ob_pd, x->x_rcv);
}

void vu_setup(void)
{
    vu_class = class_new(gensym("vu"), (t_newmethod)vu_new, (t_method)vu_free,
			 sizeof(t_vu), 0, A_GIMME, 0);
    class_addfloat(vu_class,vu_float);
    class_addmethod(vu_class, (t_method)vu_ft1, gensym("ft1"), A_FLOAT, 0);
    class_addlist(vu_class, vu_list);
    class_addmethod(vu_class, (t_method)vu_dialog, gensym("dialog"),
		    A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_size, gensym("size"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_scale, gensym("scale"), A_DEFFLOAT, 0);
    class_addmethod(vu_class, (t_method)vu_delta, gensym("delta"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_pos, gensym("pos"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_color, gensym("color"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_receive, gensym("receive"), A_DEFSYM, 0);
    class_addmethod(vu_class, (t_method)vu_label, gensym("label"), A_DEFSYM, 0);
    class_addmethod(vu_class, (t_method)vu_label_pos, gensym("label_pos"), A_GIMME, 0);
    class_addmethod(vu_class, (t_method)vu_label_font, gensym("label_font"), A_GIMME, 0);
    if(!iem_gui_key_sym)
	iem_gui_key_sym = gensym("#keyname");
    vu_widgetbehavior.w_getrectfn =    vu_getrect;
    vu_widgetbehavior.w_displacefn =   vu_displace;
    vu_widgetbehavior.w_selectfn =     vu_select;
    vu_widgetbehavior.w_activatefn =   NULL;
    vu_widgetbehavior.w_deletefn =     vu_delete;
    vu_widgetbehavior.w_visfn =        vu_vis;
#if (PD_VERSION_MINOR > 31)
    vu_widgetbehavior.w_clickfn =      NULL;
    vu_widgetbehavior.w_propertiesfn = vu_properties;
#endif
    vu_widgetbehavior.w_savefn =       vu_save;
    class_setwidget(vu_class,&vu_widgetbehavior);
    class_sethelpsymbol(vu_class, gensym("vu"));
}
